/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *  
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@austin.rr.com>. Portions created by Hollis Blanchard are Copyright
 * (C) 2001 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * Stephan Allner <allner@informatik.hu-berlin.de>
 * Gordon Heydon <gjheydon@bigfoot.com>
 * Damien Burke <damienb@mit.edu>
 * Jochen Garcke <jochen@pms.garcke.de>
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "c-client/c-client.h"
#include "c-client/smtp.h"

#include "cclientcalls.h"
#include "options.h"
#include "pilot-mailsync.h"
#include "llist.h"
#include "util.h"
#define TRUE (1)
#define FALSE (0)

#ifdef DEBUG
static void printMail(const struct Mail *mail);
#endif

static char *getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr);
static void getMsg(MailCategory mailCtg, MAILSTREAM *stream, unsigned long number);
extern void mail_string_init (STRING *s,void *data,unsigned long size);

/* everything that we got back from a search */
static listItem *gSearchResults;

/* helper functions to create/destroy c-client STRINGLIST's */
static STRINGLIST *
mkSL(char *str)
{
	STRINGLIST *sl = (STRINGLIST *)malloc(sizeof(STRINGLIST));
	sl->text.data = strdup(str);
	sl->text.size = strlen(str);
	sl->next = NULL;
	return sl;
}

void
delSL(STRINGLIST *sl)
{
	STRINGLIST *d;
	while (sl != NULL){
		free(sl->text.data);
		d = sl;
		sl = sl->next;
		free(d);
	}
}

/*
 *  Create a unique message id
 *
 *  Return: message id string
 *
 *  Uniqueness is gaurenteed by using the host name, process id, date to the
 *  second and a single unique character
 *  
 *  Basic routine taken from pine
 */
char *create_message_id(char *hostname)
{
    static short osec = 0, cnt = 0;
    char        *id;
    time_t       now;
    struct tm   *now_x;

    now   = time((time_t *)0);
    now_x = localtime(&now);
    id    = (char *)fs_get(128 * sizeof(char));

    if(now_x->tm_sec == osec){
        cnt++;
    }else{
        cnt = 0;
        osec = now_x->tm_sec;
    }
    sprintf(id,"<pms.%.20s.%02d%02d%02d%02d%02d%02d%X.%d@%.50s>",
            VERSION, (now_x->tm_year) % 100, now_x->tm_mon + 1,
            now_x->tm_mday, now_x->tm_hour, now_x->tm_min, now_x->tm_sec, 
            cnt, getpid(), hostname);

    return(id);
}

/* Output RFC 822 message
 * Accepts: temporary buffer
 *          envelope
 *          body
 *          I/O routine
 *          stream for I/O routine
 *          non-zero if 8-bit output desired
 * Returns: T if successful, NIL if failure
 *
 * Adapted out of c-client
 */

long pms_rfc822_output (char *t,ENVELOPE *env,BODY *body,soutr_t f,void *s,
			long ok8bit)
{
	char *st; 
	if (ok8bit) rfc822_encode_body_8bit (env,body);
	else rfc822_encode_body_7bit (env,body);
	rfc822_header (t,env,body);   /* build RFC822 header */
	st = &(t[strlen (t)]);
	st-=2;
	sprintf (st,"X-Mailer: %s-%s\015\012",PACKAGE,VERSION);
	/* output header and body */
	return (*f) (s,t) && (body ? rfc822_output_body (body,f,s) : T);
}

/* SMTPMail()
 *
 * send mail using stmp, called by PalmToMail() (pilot-mailsync.c)
 */
int
SMTPMail(struct Mail *mail, const char *from, const char *charset)
{
	ENVELOPE *env;
	BODY *body;
	char *hostlist[2];
	SENDSTREAM *stream;

	if (gSMTPPop){
		MAILSTREAM *popStream = NIL;
		long options = 0;
		char popName[MAILTMPLEN];
		
		sprintf(popName, "{%s/pop3}", gSMTPServer);
		mail_parameters(popStream, SET_RSHTIMEOUT, 0); /* disable RSH attempts */
		if (NULL == (popStream = mail_open(popStream, popName, options))) {
			mm_log("fatal error: couldn't open mailbox", ERROR);
			return 1;
		}
		mail_close_full(popStream, 0);
	}

	hostlist[0] = gSMTPServer;
	hostlist[1] = NIL;
	stream = smtp_open(hostlist,0);

	if (stream != 0){
		char date[MAILTMPLEN], *tmpAlloc;
		ADDRESS *fromAdd = 0, *toAdd =0, *ccAdd = 0, *bccAdd = 0, *replyAdd = 0;
		PARAMETER *parameter;

		mail_parameters(NULL, SET_RFC822OUTPUT, (void *)pms_rfc822_output);

		env = mail_newenvelope();		

		rfc822_date (date);
		env->date = cpystr(date);

		rfc822_parse_adrlist(&fromAdd, cpystr(from), 0);
		env->from = fromAdd;

		env->message_id = create_message_id(env->from->host);

		rfc822_parse_adrlist(&replyAdd, cpystr(mail->replyTo), 0);
		env->reply_to = replyAdd;

		env->subject = mail->subject ? cpystr(mail->subject) : cpystr("<none>");

		rfc822_parse_adrlist(&toAdd, cpystr(mail->to), 0);
		env->to = toAdd;

		rfc822_parse_adrlist(&ccAdd, cpystr(mail->cc), 0);
		env->cc = ccAdd;

		rfc822_parse_adrlist(&bccAdd, cpystr(mail->bcc), 0);
		env->bcc = bccAdd;

		body = mail_newbody();			
		body->type = TYPETEXT;
		

		tmpAlloc = (char *)malloc(strlen(mail->body)+
					  strlen(mail->signature ? gSig.signature : "")+MAILTMPLEN);
		strcpy(tmpAlloc,mail->body);
		strcpy(tmpAlloc+strlen(mail->body),"\n--\n");
		if (mail->signature)
			strcpy(tmpAlloc+strlen(tmpAlloc),gSig.signature);
		lf2crlf((char **)&(body->contents.text.data), tmpAlloc);
		body->contents.text.size = strlen(body->contents.text.data);
		body->encoding = ENCQUOTEDPRINTABLE;

		parameter = mail_newbody_parameter();

		parameter->attribute = cpystr("charset");
		sprintf (tmpAlloc,"%s",charset);
		parameter->value = tmpAlloc;
		body->parameter = parameter;
		
		if (smtp_mail (stream,"MAIL",env,body)){
			mail_free_body (&body);
			mail_free_envelope (&env);
			smtp_close(stream);
			return 0;
		}
		else {
			mail_free_body (&body);
			mail_free_envelope (&env);
			fprintf(stderr, "ERROR: %s returned %s\n", gSMTPServer, stream->reply);
			smtp_close(stream);
			return 1;
		}
	}
	return 2;
}

/* MailToString()
 *
 * converts a PalmMail to a (char *) with CRLF instead of LF ( :( )
 *
 * caller is responsible for freeing memory.
 */
char *
MailToString(char *mailstring, const struct Mail *mail, const char *from,
				const char *date, const char *charset)
{
	char *line=NULL, *crlf_string=NULL;
	int i;

	/* mailtab represents each line of the mail header. each non-empty line
	 * will be added to the mail.
	 */
	const char *mailtab[][2] = 	{
		{ "From: %s\r\n", from },
		{ "To: %s\r\n", mail->to },
		{ "Cc: %s\r\n", mail->cc },
		{ "Bcc: %s\r\n", mail->bcc },
		{ "Reply-To: %s\r\n", mail->replyTo },
		{ "Subject: %s\r\n", mail->subject ? mail->subject : "<none>" },
		{ "X-Mailer: %s\r\n", PACKAGE "-" VERSION},
		{ "Content-Type: text/plain; charset=%s\r\n", charset},
		{ "Date: %s\r\n", date },
		{ "\r\n%s\r\n", mail->body },
		{ "-- \r\n%s\r\n", mail->signature ? gSig.signature : "" }
	};

	for (i=0; i<11; i++) {
		if (mailtab[i][1] && strlen(mailtab[i][1])) {
			/* make a buffer for this line */
			line = (char *)realloc(line, strlen(mailtab[i][0]) +
					       strlen(mailtab[i][1]) + 1);
			sprintf(line, mailtab[i][0], mailtab[i][1]);

			/* Make sure the \r\n is not duplicated at the end of
			 * the string. If it is, strip off the last two chars
			 */
			crlf_string = strstr(line, "\n\r\n");
			if ( crlf_string != NULL && *(crlf_string +3) == '\0') {
				*crlf_string = '\0';    /* strip last three*/
				strcat(line, "\r\n");   /* add cr+lf+null */
			}
			crlf_string = NULL;

			/* make sure there's not a final \r (^M) */
			if (index(line,'\r') != NULL) {
				if (rindex(line,'\r') +1 == index(line, '\0') ) {
					*rindex(line, '\r') = '\n';
				}
			}

			/* resize the mailstring to fit the new line */
			mailstring = (char *)realloc(mailstring, strlen(mailstring) +
											strlen(line)+1);

			/* append this line to the string */
			strcat(mailstring, line);
		}
	}
	free(line);

	/* take our hard-earned string and make sure it contains only CRLF's */
	lf2crlf(&crlf_string, mailstring);
	mailstring = (char *)realloc(mailstring, strlen(crlf_string)+1);
	strncpy(mailstring, crlf_string, strlen(crlf_string)+1);
	free(crlf_string);

	return mailstring;
}

/* AddMailToMailbox()
 *
 * Adds a PalmMail to the specified mailbox
 */
int
AddMailToMailbox(char *mailbox, struct Mail *mail, const char *from) {
	STRING msg;
	char *mailstring;
	char date[MAILTMPLEN], flags[MAILTMPLEN];

	rfc822_date(date);
	strcpy(flags, "(\\Seen)"); /* mark mail as read */

	mailstring = (char *)malloc(1);
	mailstring[0] = '\0';
	mailstring = MailToString(mailstring, mail, from, date, gCharSet);

	/* write out the standard email headers */
	INIT(&msg, mail_string, (void *)mailstring, strlen(mailstring));
	mail_append_full(NIL, mailbox, flags, date, &msg);

	free(mailstring);
	return 0;
}

/* OpenMail()
 * 
 * opens connection to the remote mailbox
 */
MAILSTREAM *
OpenMail(char *mailbox)
{
	MAILSTREAM *stream = NIL;
	long options = 0;

/********* include a .c file... **********/
#include "c-client/linkage.c"
/*****************************************/

#ifdef DEBUG
	options = OP_DEBUG;
#endif

	mail_parameters(stream, SET_RSHTIMEOUT, 0); /* disable RSH attempts */
	if (NULL == (stream = mail_open(stream, mailbox, options))) {
		mm_log("fatal error: couldn't open mailbox", ERROR);
		return stream;
	}
	return stream;
}

/* CloseMail()
 *
 * closes the mail connection
 */
void
CloseMail(MAILSTREAM *mailStream)
{
	mail_close_full(mailStream, CL_EXPUNGE); /* expunge deleted mails */
}


/* MailToPalm()
 *
 * transfer mail from desktop to Palm
 */
void
MailToPalm(MailCategory mailCtg, MAILSTREAM *mailStream)
{
	SEARCHPGM search, notsearch, imptsearch, moresearch;
	SEARCHPGM *curpgm=&search;
	SEARCHPGMLIST notlist;
	SEARCHOR orsearch;
	int msgno;

	memset(&search, 0, sizeof(SEARCHPGM));
	memset(&imptsearch, 0, sizeof(SEARCHPGM)); /* get flagged msgs */
	memset(&moresearch, 0, sizeof(SEARCHPGM)); /* complement to imptsearch */
	memset(&notsearch, 0, sizeof(SEARCHPGM));

	memset(&orsearch, 0, sizeof(SEARCHOR));
	memset(&notlist, 0, sizeof(SEARCHPGMLIST));

	if (gPrefs.syncType == mailSyncUnread) {
		if (gSyncSeen) {
			/* if we are updating the seen flag later, we need to only get
			 * recent messages (otherwise we'll get duplicates)
			 */
			search.recent = 1;
		} else {
			/* search for new messages */
			search.unseen = 1;
		}

	} else if (gPrefs.syncType == mailSyncAll){
		/* do nothing, all is done already */
		/* Hmm, one maybe could move SynchronizeFlags from 
		   pilot-mailsync.c to here */
	} else if (gPrefs.syncType == mailSyncFilter) {
		/* set up some filters */

		if (gPrefs.getHigh) {
			/* get all Important msgs, regardless of include/exclude */
			search.or = &orsearch;
			orsearch.first = &imptsearch; /* Important messages OR ... */
			orsearch.second = &moresearch;
			imptsearch.flagged = 1;
			curpgm = &moresearch; /* future filter settings go to or.second */
		}

		if (gPrefs.getContaining) {
			/* "include" search */
			if (gPrefs.filterTo)
				curpgm->to = mkSL(gPrefs.filterTo);
			if (gPrefs.filterFrom)
				curpgm->from = mkSL(gPrefs.filterFrom);
			if (gPrefs.filterSubject)
				curpgm->subject = mkSL(gPrefs.filterSubject);
		} else {
			/* "exclude" search */
			curpgm->not = &notlist;
			notlist.pgm = &notsearch;

			if (gPrefs.filterTo)
				notsearch.to = mkSL(gPrefs.filterTo);
			if (gPrefs.filterFrom)
				notsearch.from = mkSL(gPrefs.filterFrom);
			if (gPrefs.filterSubject)
				notsearch.subject = mkSL(gPrefs.filterSubject);
		}
	}

	gSearchResults = NULL;
	/* this stores all messages in gSearchResults via mm_searched() */
	mail_search_full(mailStream, gCharSet, &search, NIL);

	/* now process the messages that matched */
	while ((msgno = (int)pop(&gSearchResults)))
		getMsg(mailCtg, mailStream, msgno);
	infoOutput("\n"); /* terminate "...." from getMsg */

	/* free mkSL memory */
	delSL(curpgm->to);
	delSL(curpgm->from);
	delSL(curpgm->subject);
	delSL(notsearch.to);
	delSL(notsearch.from);
	delSL(notsearch.subject);
}

/* getMsg
 *
 * do the actual message processing:
 * - get from c-client (delete if appropriate)
 * - construct new Palm Mail item
 * - transfer to Palm
 */
void
getMsg(MailCategory mailCtg, MAILSTREAM *stream, unsigned long number)
{
	BODY *body;
	MESSAGECACHE *elt;

	mail_fetchstructure(stream, number, &body);

	if (body) {
		/* 8000 bytes is the max truncation size, plus 500 bytes for to, cc,
		 * from, and subject lines
		 */
		unsigned char record[8000 + 500];
		struct Mail pilotMail;
		char *datestr;
		unsigned long len=0;

		/* mark mail to be deleted */
		if (gDelete && (gPrefs.syncType != mailSyncUnread)) {
			/* Palm Mail app says "all messages will remain on the E-Mail
			 * application" in sync mode Unread.
			 */
			markDeleted(number, stream);
		}

		/* I don't know what the following do for Palm-bound mail */
		pilotMail.signature = 0;
		pilotMail.confirmRead = 0;
		pilotMail.confirmDelivery = 0;
		pilotMail.addressing = 0;

		elt = mail_elt(stream, number); /* for msg flags */

		if (elt->spare){
			infoOutput("o");
			return;
		}

		/* in the Palm OS Mail struct, 0 is high priority (presumably 2 is low,
		 * but low doesn't seem to be visibly marked in the Mail app)
		 */
		pilotMail.priority = !(elt->flagged);
		pilotMail.read = elt->seen;

		/* convert message's text-date to struct tm */
		datestr = strip_cr(getHeader(stream, number, "Date"));
		if (datestr != 0){
			pilotMail.date = *DateToTm(datestr);
			free(datestr);
			pilotMail.dated = 1;
		}
		/* fill in all the text headers */
		pilotMail.subject = strip_cr(getHeader(stream, number, "Subject"));
		pilotMail.from = strip_cr(getHeader(stream, number, "From"));
		pilotMail.to = strip_cr(getHeader(stream, number, "To"));
		pilotMail.cc = strip_cr(getHeader(stream, number, "Cc"));
		pilotMail.bcc = strip_cr(getHeader(stream, number, "Bcc"));
		pilotMail.replyTo = strip_cr(getHeader(stream, number, "Reply-To"));
		pilotMail.sentTo = strip_cr(getHeader(stream, number, "Sent-To"));

		/* get body, truncating if necessary. Note that fetchtext_full
		 * should return the entire body without seperating parts
		 */
		pilotMail.body = strip_cr(mail_fetchtext_full(stream, number,
							      &len,
							      /* gSyncSeen ? FT_PEEK : NIL*/
							      FT_PEEK)
					  );
		if (len > gPrefs.truncate) {
			/* write over the last few chars of the mail */
			char *trunc_warning = "\n\n[pilot-mailsync truncated]";
			strcpy(pilotMail.body+gPrefs.truncate-strlen(trunc_warning),
					trunc_warning);
			infoOutput("/");
			if (gVerbose)
				printf("\n  truncating mail #%li; new len is %i\n", number,
						strlen(pilotMail.body));
		}

#ifdef DEBUG
			printMail(&pilotMail);
#endif

		/* finished setting all of the headers; write it to the record buffer */
		len = pack_Mail(&pilotMail, record, sizeof(record));

		/* write the mail record to the Palm now */
		if (0 < dlp_WriteRecord(gSocket, gDB, 0, 0, mailCtg, record, len, 0)) {
			/* yay, wrote mail to Palm */
			nToPalm++;
			infoOutput("."); /* some feedback - one '.' per delivered mail */
		} else {
			mm_log("  failure writing to Palm\n", ERROR);
		}
	}
}

/* makeSearchItem()
 * 
 * in text to search for
 * out stringlist to search for
 *
 * if text includes mimestuff convert it to utf8
 *
 * returns 1 if utf8, 0 otherwise
 */
int makeSearchItem(STRINGLIST **dstSL, char *text)
{
/* if mail source doesn't support utf-8 
 * configure with --disable-utf8 (option to be included in configure)
 */ 
#ifdef _NOT_USE_UTF_8_ 
	char *dummy = strip_cr_lf(text);		
	*dstSL = mkSL(dummy);
	free(dummy);
	return 0;
#else /* default config */
#include "c-client/utf8.h"
	SIZEDTEXT a,b;
	char text2[MAILTMPLEN];
	a.data = text;
	a.size = strlen(text);
	b.data = text2;
	if (!utf8_mime2text(&a,&b))
		mm_log ("Error in utf8_mime2text",ERROR);
	if (strcmp(text,b.data)){
		/* convert to utf8 changed the string */
		char *dummy = strip_cr_lf(b.data);
		*dstSL = mkSL(dummy);
		free(dummy);
		return TRUE;
	}
	else{
		char *dummy = strip_cr_lf(text);		
		*dstSL = mkSL(dummy);
		free(dummy);
		return FALSE;
	}
#endif
}

/* makeSearchList()
 * 
 * in text to search for
 * out stringlist to search for
 *
 * if text includes mimestuff convert it to utf8
 *
 * returns 1 if utf8, 0 otherwise
 */
int makeSearchList(STRINGLIST **dstSL, char *inputText, int internalSearch)
{
	int i=0, quote=FALSE, utf_flag =FALSE;
	char *p1;
	char text[MAILTMPLEN];
	STRINGLIST *buffer = NULL;
	p1 = inputText;
	while (*p1 != '\0') {
		if (quote || *p1 != ','){
			if (*p1 == '"'){
				quote = quote ? FALSE : TRUE;
			}
			if (  !(*p1 == '"' && internalSearch) ){
				text[i] = *p1;
				i++;
			}
			p1++;
		}
		else{
			text[i] = '\0';
			if (makeSearchItem(dstSL, text))
				utf_flag = 1;
			if (buffer != NULL)
				(*dstSL)->next = buffer;
			buffer = *dstSL;
			i=0;
			p1++;
			while ( (*p1 == ' ') || (*p1 == '\r') || ( *p1 == '\n')) {
				p1++;
			}
		}
	}
	text[i] = '\0';
	if (makeSearchItem(dstSL, text))
		utf_flag = TRUE;
	if (buffer != NULL)
		(*dstSL)->next = buffer;
	return utf_flag;
}

/* findMail()
 *
 * find (all) message(s) matching *pilotMail
 * caller is responsible for clearing gSearchResults
 */
listItem *
findMail(struct Mail *pilotMail, MAILSTREAM *mailStream)
{
	SEARCHPGM search;
	int utf_flag = FALSE, internalSearch = FALSE;

	gSearchResults = NULL;
	memset(&search, 0, sizeof(SEARCHPGM));
	
	/* Are we using internal c-client search */
	if (mailStream->dtb->search == NULL)
		internalSearch = TRUE;

	if (pilotMail->subject){
		if (makeSearchItem(&(search.subject), pilotMail->subject))
			utf_flag = TRUE;
	}

	if (pilotMail->from)
		if (makeSearchList(&(search.from), pilotMail->from, internalSearch))
			utf_flag = TRUE;

	if (pilotMail->to){
		if (makeSearchList(&(search.to), pilotMail->to, internalSearch))
			utf_flag = TRUE;
	}

	if (pilotMail->cc)
		if (makeSearchList(&(search.cc), pilotMail->cc, internalSearch))
			utf_flag = TRUE;

	if (pilotMail->bcc)
		if (makeSearchList(&(search.bcc), pilotMail->bcc, internalSearch))
			utf_flag = TRUE;

	/* 1900 is baseyear of struct tm */

	search.senton = (  (pilotMail->date.tm_year - BASEYEAR + 1900) << 9) +
		((pilotMail->date.tm_mon+1) << 5) +
		pilotMail->date.tm_mday;

	if (utf_flag)
		mail_search_full(mailStream, "UTF-8", &search, NIL);
	else
		mail_search_full(mailStream, gCharSet, &search, NIL);

	if (search.subject)
		delSL(search.subject);
	if (search.from)
		delSL(search.from);
	if (search.to)
		delSL(search.to);
	if (search.cc)
		delSL(search.cc);
	if (search.bcc)
		delSL(search.bcc);

	/* post-processing of gSearchResults if more than one mail matches
	 * use time and sec to filter out mails which don't really correspond
	 */

	if (gSearchResults!= 0 && gSearchResults->next != 0){
		listItem *current;
		current = gSearchResults;
		
		while (current){
			struct tm *mailTime;
			ENVELOPE *env;
			env = mail_fetchstructure(mailStream, (int)(current->data), NULL);
			mailTime = DateToTm(env->date);		       
			if (mailTime->tm_hour != pilotMail->date.tm_hour || mailTime->tm_min != pilotMail->date.tm_min)
				removeLI(&current, &gSearchResults);
			else
				current = current->next;
		}
	}
	
	return gSearchResults;
}

/* markDeleted
 *
 * mark message numbered 'msgno' as Deleted on the server
 */
void
markDeleted(unsigned long msgno, MAILSTREAM *mailStream)
{
	char seq[5]; /* won't handle >9,999 msg mailboxes */

	sprintf(seq, "%li", msgno);

#ifdef DEBUG
	printf("  marking mail #%s for deletion\n", seq);
#endif

	mail_setflag(mailStream, seq, "\\DELETED");
}

/* markSeen
 *
 * mark message numbered 'msgno' as Seen on the server
 */
void
markSeen(unsigned long msgno, MAILSTREAM *mailStream)
{
	char seq[5]; /* won't handle >9,999 msg mailboxes */

	sprintf(seq, "%li", msgno);

#ifdef DEBUG
	printf("  marking mail #%s as seen\n", seq);
#endif

	mail_setflag(mailStream, seq, "\\SEEN");
}

/* markSeenAndOnPalm
 *
 * mark message numbered 'msgno' as Seen on the server
 *
 * spare = 1 is set for marking MailIsOnPalm
 */
void
markSeenAndOnPalm(unsigned long msgno, MAILSTREAM *mailStream)
{
	BODY *body;
	MESSAGECACHE *elt;
	char seq[5]; /* won't handle >9,999 msg mailboxes */

	sprintf(seq, "%li", msgno);

#ifdef DEBUG
	printf("  marking mail #%s as seen\n", seq);
#endif

	mail_setflag(mailStream, seq, "\\SEEN");

	mail_fetchstructure(mailStream, msgno, &body);
	elt = mail_elt(mailStream, msgno);
	elt->spare = 1;
}

/* isSeen
 *
 * return true if message number 'msgno' is marked Seen on the server
 * return false otherwise
 */
int
isSeen(unsigned long msgno, MAILSTREAM *mailStream)
{
	BODY *body;
	MESSAGECACHE *elt;

	mail_fetchstructure(mailStream, msgno, &body);
 
	if (body) {
		elt = mail_elt(mailStream, msgno);
		return elt->seen;
	}
	return 0;
}

/* isSeenMarkOnPalm
 *
 * return true if message number 'msgno' is marked Seen on the server
 * return false otherwise
 *
 * spare = 1 is set for marking MailIsOnPalm
 */
int
isSeenMarkOnPalm(unsigned long msgno, MAILSTREAM *mailStream)
{
	BODY *body;
	MESSAGECACHE *elt;

	mail_fetchstructure(mailStream, msgno, &body);
 
	if (body) {
		elt = mail_elt(mailStream, msgno);
		elt->spare = 1;
		return elt->seen;
	}
	return 0;
}

/* mm_searched: called on every message that matches the conditions set in
 *              mail_search_full (above).
 * 
 */
void
mm_searched (MAILSTREAM *stream, unsigned long number)
{
	push(&gSearchResults, (void *)number);
}

void
mm_login(NETMBX *mb, char *user, char *pwd, long trial)
{
	char *nl=NULL;

#ifdef DEBUG
	printf("acquiring login info\n");
#endif
	if (trial || *gUser == '\0') {
		printf("login: ");
		fgets(gUser, NETMAXUSER, stdin);
		/* get rid of the trailing \n */
		if ((nl=strchr(gUser, '\n')))
			*nl = '\0';
	}

	if (trial || *gPwd == '\0')
		strncpy(gPwd, getpass("Password: "), MAILTMPLEN);

	strncpy(user, gUser, NETMAXUSER);
	strncpy(pwd, gPwd, MAILTMPLEN);
}

void
mm_log (char *string,long errflg)
{
	switch ((short) errflg) {
		case NIL:
			if (gVerbose)
				printf ("-- %s\n", string);
			break;
		case PARSE:
		case WARN:
			printf ("== Warning: %s\n", string);
			break;
		case ERROR:
			printf ("** Error: %s\n", string);
			break;
	}
}

#ifdef DEBUG
void
printMail(const struct Mail *mail)
{
	if (mail->subject) printf("  subject: %s", mail->subject);
	if (mail->from) printf("  from: %s", mail->from);
	if (mail->to) printf("  to: %s", mail->to);
	if (mail->cc) printf("  cc: %s", mail->cc);
	if (mail->bcc) printf("  bcc: %s", mail->bcc);
	if (mail->replyTo) printf("  replyTo: %s", mail->replyTo);
	if (mail->sentTo) printf("  sentTo: %s", mail->sentTo);
}
#endif

char *
getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr)
{
	STRINGLIST *header_list;
	char *hdr=NULL;

	if (NULL == (header_list = mail_newstringlist())) {
		mm_log("Ran out of memory in getHeader()\n", ERROR);
		return NULL;
	}

	header_list->text.data = (unsigned char *)strdup(headerstr);
	header_list->text.size = strlen((char *)header_list->text.data);

	/* skip 'headerstr: ' */
	hdr = strchr(mail_fetchheader_full(stream,msg_no,header_list,NIL,NIL), ' ');
	if (hdr) {
		hdr++; /* skip leading space */
		/* convert \r\n\r\n to \n\0 */
		hdr[strlen(hdr)-4] = '\n'; 
		hdr[strlen(hdr)-3] = '\0'; 
	}

	free(header_list->text.data);
	return hdr;
}


/***** nothing to see below this point *****/

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
	mm_log(string, errflg);
}

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
	abort();
	return NIL;
}

void mm_fatal (char *string)
{
	mm_log(string, ERROR);
}

void mm_dlog (char *string)
{
	puts(string);
}

void mm_critical (MAILSTREAM *stream)
{
}

void mm_nocritical (MAILSTREAM *stream)
{
}

void smtptest (long debug)
{
}

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
}

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
}

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
}

void mm_list (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_lsub (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
}

