/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@austin.rr.com>. Portions created by Hollis Blanchard are Copyright
 * (C) 2001 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * Joachim Reichelt <Reichelt@gbf.de
 * 
 */


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <sys/types.h>

#include "c-client/c-client.h"

#include "options.h"
#include "pilot-mailsync.h"
#include "util.h"

#define FILEMAX 128
#ifndef MAXHOSTNAMELEN
#	define MAXHOSTNAMELEN 64
#endif

char gConfigFile[MAILTMPLEN];
char gDevice[MAILTMPLEN];
char gSendmail[MAILTMPLEN];
char gMailbox[MAILTMPLEN];
char gOutbox[MAILTMPLEN];
char gDrafts[MAILTMPLEN];
char gFiled[MAILTMPLEN];
char gPwd[MAILTMPLEN];
char gFrom[MAILTMPLEN];
char gReplyTo[MAILTMPLEN];
char gUser[NETMAXUSER];
char gCharSet[MAILTMPLEN];
char gSMTPServer[MAXHOSTNAMELEN];
int gSMTPPort;
int gSMTPPop;
int gDelete;
int gDeleteRead;
int gSyncSeen;
int gVerbose;
int gSendFileOnly;
/* c-client can't tell if POP3 or MH mails are flagged "new", so all mails are
 * marked as "new".
 *
 * To work around this (especially with delete, since all "new" mails could be
 * deleted after transfer to Palm), we can:
 * 1. make a local copy of every mail currently in the MailDB, and
 * 2. before copying a message to the Palm, check to see if we already have a
 *    local copy.
 *
 * This is slow with large Palm MailDB's, and isn't perfect because if a mail
 * is deleted from the Palm but not from the mail source, on the next sync it
 * will be copied back to the Palm. This is considered an unfortunate hack.
 */
int gLocalHack;


/* private */
static void load_defaults(void);
static void load_configfile(void);
static void load_env(void);
static void load_cmdline(int argc, char *argv[]);
static int check_opts(void);
static void Help(char *prog_name, int longhelp);



void load_options(int argc, char *argv[], int cmdlineFlag)
{
	load_defaults();
	load_env();
	load_configfile();
	if (cmdlineFlag)
		load_cmdline(argc, argv);
	if (check_opts()) {
		fprintf(stderr, "fatal config error; exiting\n");
		exit(1);
	}
	if (strstr(gMailbox, "/pop3") || strstr(gMailbox, "#mh"))
		gLocalHack = 1;
}

void load_defaults(void)
{

	char hn[MAXHOSTNAMELEN];
	char *username;
	struct passwd *pwd;

	strcpy(gConfigFile, ".pilotmailrc");
	strcpy(gDevice, "/dev/pilot");
	strcpy(gSendmail, "/usr/sbin/sendmail -t -i");
	strcpy(gMailbox, "Mailbox");
	strcpy(gOutbox, "");
	strcpy(gDrafts, "");
	strcpy(gFiled, "");
	strcpy(gUser, "");
	strcpy(gPwd, "");
	strcpy(gSMTPServer, "");
	/* Maybe get gCharSet via nl_langinfo */
	strcpy(gCharSet, "US-ASCII");
	strcpy(gReplyTo, "");

	/* set default From address to Realname <$USER@$HOSTNAME> */
	gethostname(hn,MAXHOSTNAMELEN-1);

	pwd = getpwuid(getuid());
	username = pwd->pw_name;

	if (pwd && pwd->pw_gecos){
		if (strlen(username) + strlen(hn) + strlen(pwd->pw_gecos) + 3 > MAILTMPLEN-1) {
			printf( "WARNING: \"%s <%s@%s>\" is longer than %i chars!\n"
				"Please report this to the pilot-mailsync maintainer.\n",
				pwd->pw_gecos, username, hn, MAILTMPLEN);
			return;
		}
		strncpy(gFrom, pwd->pw_gecos, MAILTMPLEN-1-strlen(pwd->pw_gecos));
		strcat(gFrom," <");
		strncat(gFrom, username,
			MAILTMPLEN-1-strlen(gFrom));
	}
	else{
		if (strlen(username) + strlen(hn) > MAILTMPLEN-1) {
			printf( "WARNING: \"%s@%s\" is longer than %i chars!\n"
				"Please report this to the pilot-mailsync maintainer.\n",
				username, hn, MAILTMPLEN);
			return;
		} 
		strncpy(gFrom, username,
			MAILTMPLEN-1-strlen(username));
	}
	strcat(gFrom, "@");
	strncat(gFrom, hn, MAILTMPLEN-1-strlen(gFrom));
	if (pwd && pwd->pw_gecos)
		strcat(gFrom,">");
	
	gSMTPPort = 25;
	gSMTPPop = 0;
	gDelete = 0;
	gDeleteRead = 0;
	gSyncSeen = 0;
	gVerbose = 0;
	gSendFileOnly = 0;
	endpwent();
}

void load_env(void)
{
	char *envstr;

#ifdef DEBUG
	printf("loading environment variables\n");
#endif

	if ((envstr = getenv("PILOTPORT")))
		strncpy(gDevice, envstr, MAILTMPLEN);

	if ((envstr = getenv("MAIL")))
		strncpy(gMailbox, envstr, MAILTMPLEN);

	if ((envstr = getenv("PILOTFROM")))
		strncpy(gFrom, envstr, MAILTMPLEN);

	if ((envstr = getenv("REPLYTO")))
	  strncpy(gReplyTo, envstr, MAILTMPLEN);

	if ((envstr = getenv("LOGINNAME")))
		strncpy(gUser, envstr, NETMAXUSER);
}

/* XXX maybe load_configfile(char *file) later? */
void load_configfile(void)
{
	char *home = getenv("HOME");
	char filepath[FILEMAX]={0};
	FILE *cfgfile = 0;
	const int linesize=128;
	char line[linesize];

	strcat(filepath, home);
	strcat(filepath, "/");
	strcat(filepath, gConfigFile);

#ifdef DEBUG
	printf("loading config file %s\n", filepath);
#endif

	if (NULL == (cfgfile = fopen(filepath, "r"))) {
		if (gVerbose)
			printf("  could not open %s\n", filepath);
		return;
	}

	/*
	 * read a line:
	 * 'variable=value' or 'option'
	 */
	while ((char*)NULL != fgets(line, linesize, cfgfile)) {
		char *value=line;
		char *var=line;
		int boundchk=0;

		line[strlen(line)-1] = '\0'; /* strip \n */

		if ((var[0] == '#') || (var[0] == '\0'))
			continue; /* skip comment and blank lines */

		while ((line[boundchk] != '=') && (boundchk < linesize-1))
			boundchk++; /* skip var */
		line[boundchk++] = '\0'; /* terminate var at the '=' */

		/* now boundchk points to the beginning of the 2nd word (if any) */
		value = &line[boundchk];

#ifdef DEBUG
		printf("  '%s'='%s'\n", var, value);
#endif
		if (!strcmp(var, "device")) {
			strncpy(gDevice, value, MAILTMPLEN);

		} else if (!strcmp(var, "sendmail")) {
			strncpy(gSendmail, value, MAILTMPLEN);

		} else if (!strcmp(var, "user")) {
			strncpy(gUser, value, NETMAXUSER);

		} else if (!strcmp(var, "from")) {
			strncpy(gFrom, value, MAILTMPLEN);

		} else if (!strcmp(var, "reply-to")) {
			strncpy(gReplyTo, value, MAILTMPLEN);

		} else if (!strcmp(var, "mailbox")) {
			strncpy(gMailbox, value, MAILTMPLEN);

		} else if (!strcmp(var, "outbox")) {
			strncpy(gOutbox, value, MAILTMPLEN);

		} else if (!strcmp(var, "drafts")) {
			strncpy(gDrafts, value, MAILTMPLEN);

		} else if (!strcmp(var, "filed")) {
			strncpy(gFiled, value, MAILTMPLEN);

		} else if (!strcmp(var, "passwd")) {
			/* bitch at user if permissions are uncool */
			struct stat fileinfo;
			if (-1 != stat(filepath, &fileinfo)) {
				if ((fileinfo.st_mode & 0777) ^ (S_IRUSR|S_IWUSR)) {
					fprintf(stderr,
							"%s contains \"passwd\" but is insecure!\n"
							"Please 'chmod 600 %s'.\n",
							filepath, filepath);
					exit(1);
				}
			}
			strncpy(gPwd, value, MAILTMPLEN);

		} else if (!strcmp(var, "deleteread")) {
			if (!strcmp(value, "on"))
				gDeleteRead = 1;
			else
				gDeleteRead = 0;

		} else if (!strcmp(var, "delete")) {
			if (!strcmp(value, "on"))
				gDelete = 1;
			else
				gDelete = 0;

		} else if (!strcmp(var, "syncseen")) {
			if (!strcmp(value, "on"))
				gSyncSeen = 1;
			else
				gSyncSeen = 0;

		} else if (!strcmp(var, "verbose")) {
			if (!strcmp(value, "on"))
				gVerbose = 1;
			else
				gVerbose = 0;

		} else if (!strcmp(var, "smtpserver")) {
			strncpy(gSMTPServer, value, MAXHOSTNAMELEN);

		} else if (!strcmp(var, "charset")) {
			strcpy(gCharSet, value);
			
		} else if (!strcmp(var, "smtpport")) {
			gSMTPPort = atoi(value);

		} else if (!strcmp(var, "smtppop")) {
			if (!strcmp(value, "on"))
				gSMTPPop = 1;
			else
				gSMTPPop = 0;
	
		} else if (!strcmp(var, "sendfileonly")) {
			if (!strcmp(value, "on"))
				gSendFileOnly = 1;
			else
				gSendFileOnly = 0;

		}
	}
	fclose(cfgfile);
}

void load_cmdline(int argc, char *argv[])
{
	int c, errflag=0, longhelp=0;

#ifdef DEBUG
	printf("reading cmdline\n");
#endif

	while ((c = getopt(argc, argv, "vdlrp:u:f:t:s:m:?h")) != -1) {
		switch (c) {
			case 'v':
				gVerbose = 1;
				break;
			case 'd':
				gDelete = 1;
				break;
			case 'r':
				gDeleteRead = 1;
				break;
			case 'l':
				gSyncSeen = 1;
				break;
			case 'p':
				strncpy(gDevice, optarg, MAILTMPLEN);
				break;
			case 'u':
				strncpy(gUser, optarg, MAILTMPLEN);
				break;
			case 'f':
				strncpy(gFrom, optarg, MAILTMPLEN);
				break;
            		case 't':
              			strncpy(gReplyTo, optarg, MAILTMPLEN);
				break;
			case 's':
				strncpy(gSendmail, optarg, MAILTMPLEN);
				break;
			case 'm':
				strncpy(gMailbox, optarg, MAILTMPLEN);
				break;
			case '?':
			case 'h':
				errflag++;
				longhelp=1;
				break;
			default:
#ifdef DEBUG
				printf("  unknown switch %c\n", c);
#endif
				errflag++;
				break;
		}
	}

	/* too many arguments */
	if (optind != argc) {
#ifdef DEBUG
		printf("  too many arguments\n");
#endif
		errflag++;
	}

	if (errflag) {
		Help(argv[0], longhelp);
		exit(1);
	}
}

/* check for fatal errors */
int check_opts(void)
{
	int err=0;

	if (gDevice[0] == '\0') {
		fprintf(stderr, " sync with what device?");
		err++;
	} else if (gSendmail[0] == '\0' && gSMTPServer[0] == '\0') {
		fprintf(stderr, " where is sendmail?");
		err++;
	} else if (gMailbox[0] == '\0') {
		fprintf(stderr, " read from what mailbox?");
		err++;
	} else if (gFrom[0] == '\0') {
		fprintf(stderr, " use what from address?");
		err++;
	}

	/* user name and passwd can be entered in mm_login (if necessary), and so
	 * aren't checked here.
	 */

	if (gSyncSeen)
		fprintf(stderr, "The option syncseen will not be supported in the next version.\n");
#ifdef DEBUG
	printf("gDevice:     '%s'\n", gDevice);
	printf("gSendmail:   '%s'\n", gSendmail);
	printf("gUser:       '%s'\n", gUser);
	printf("gPwd:        '%s'\n", gPwd);
	printf("gMailbox:    '%s'\n", gMailbox);
	printf("gOutbox:     '%s'\n", gOutbox);
	printf("gFrom:       '%s'\n", gFrom);
	printf("gReplyTo:    '%s'\n", gFrom);
	printf("gCharSet:    '%s'\n", gCharSet);
	printf("gSMTPServer: '%s'\n", gSMTPServer);
	printf("gSMTPPort:   '%i'\n", gSMTPPort);
	printf("gSMTPPop:    '%i'\n", gSMTPPop);
	printf("gDelete:     '%s'\n", gDelete ? "on" : "off");
	printf("gDeleteRead: '%s'\n", gDeleteRead ? "on" : "off");
	printf("gSyncSeen:   '%s'\n", gSyncSeen ? "on" : "off");
	printf("gVerbose:    '%s'\n", gVerbose ? "on" : "off");
#endif

	return err;
}

void Help(char *prog_name, int longhelp)
{
	printf( "\npilot-mailsync " VERSION "\n");
	printf( "Please read docs/config for a more complete description.\n");
	printf( "usage: %s [options]\n", prog_name);
	printf( " -d                 delete transferred mail\n"
			" -r                 delete all read mail from Palm Inbox\n"
			" -l                 keep Palm and source Seen flags sync'ed\n"
			" -v                 verbose\n"
			" -p <device>        serial device connected to handheld\n"
			" -u <user>          user name for remote (POP3, IMAP) logins\n"
			" -f <from address>  from address for mail sent from handheld\n"
			" -s <sendmail cmd>  command used to send outgoing mail\n"
			" -m <mailbox>       c-client mailbox string (see docs/mailbox)\n"
			" -? -h              help\n"
			);

	if (longhelp) {
		printf( "\nrelevent environment variables:\n"
				"LOGINNAME      login name (POP and IMAP)\n"
				"MAIL           mailbox to read from (see docs/mailbox)\n"
				"PILOTFROM      'from' address\n"
				"PILOTPORT      tty device connected to PDA\n"
				);
	}

	exit(0);
}
