/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@austin.rr.com>. Portions created by Hollis Blanchard are Copyright
 * (C) 2001 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Gordon Heydon <gjheydon@bigfoot.com>
 * Jochen Garcke <jochen@pms.garcke.de>
 *
 */

#include <stdio.h>
#include <stdlib.h>

#if HAVE_SYS_WAIT_H
#	include <sys/wait.h> /* WEXITSTATUS */
#endif
#ifndef WEXITSTATUS
#	define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif

#include <pi-source.h>
#include <pi-socket.h>
#include <time.h> /* XXX pilot-link bug: required before pi-mail.h */
#include <pi-mail.h>
#include <pi-dlp.h>

#include "cclientcalls.h"
#include "palmmail.h"
#include "pilot-mailsync.h"
#include "options.h"
#include "util.h"

/* PalmToMail()
 *
 * loop over all msgs in Outbox and pass them to sendmail (via SendMail())
 */
void PalmToMail(const char *from)
{
	unsigned char buffer[0xffff]; /* XXX could probably be (a lot) smaller */
	int len, size, attr, sendmail_ret;
	recordid_t id;
	int clear_replyto_pointer_later;
	nFromPalm = 0;
	
	/* XXX pilot-link bug: does dlp_ReadNextRecInCategory not require bufsize
	 * anywhere?
	 */
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtgOutbox,
						     buffer, &id, 0, &size,
						     &attr))) {
		struct Mail mail;

		/* Skip deleted or archived records */
		if ((attr & dlpRecAttrDeleted) || (attr & dlpRecAttrArchived))
			continue;

		unpack_Mail(&mail, buffer, len);

		/*
		** Insert new Reply-To: part, but only if:
		**  a) There is no original Reply-To part
		**  b) There is a new Reply-To part  (I use two
		**     possible 'no value' strings. One is the
		**     empty string "", the other is "false")
		**
		** WATCH OUT! We cannot permit pi-mail to free
		** this part. In case we changed the replyTo field
		** we have to set it back to NULL before calling
		** free_Mail().
		*/
		if( mail.replyTo == NULL &&
		    gReplyTo[0] != '\0' &&
		    strcmp(gReplyTo, "false") != 0 )
		{
		  clear_replyto_pointer_later = 1;
		  mail.replyTo = gReplyTo;
		}
		else
		  clear_replyto_pointer_later = 0;		  

		/* for offline use:
		 * if user specified append to local file only 
		 * and do the sending later with some other tool 
		 */
		if (!gSendFileOnly)
			/* if user specified SMTP, use that. otherwise, exec sendmail */
			if (strlen(gSMTPServer)) {
				sendmail_ret = SMTPMail(&mail, from, gCharSet);
			} else {
				sendmail_ret = SendMail(&mail, from, gCharSet);
			}
		else
			sendmail_ret = 0;
			
		if (0 != sendmail_ret) {
			if (strlen(gSMTPServer)) {
				if (sendmail_ret == 2)
					/* something bad happened w/ SMTP */			
					fprintf(stderr, "ERROR: Could not connect to %s\n", gSMTPServer);
			}
			else {
				/* something bad happened w/ sendmail */			
				fprintf(stderr, "ERROR: %s returned %i\n", gSendmail,
					WEXITSTATUS(sendmail_ret));
			}
		} else {
			/* no error, so delete the just-sent mail*/
			if (strlen(gOutbox)) {
				/* save in outbox */
				AddMailToMailbox(gOutbox, &mail, from);
			}
			nFromPalm++;
			dlp_DeleteRecord(gSocket, gDB, 0, id);
		}

		/*
		** If we inserted a reply-to, we have to remove it
		** before pi-mail tries to clean it up.
		*/
		if( clear_replyto_pointer_later )
  		        mail.replyTo = NULL;

		free_Mail(&mail);
	}
}

/* SendMail()
 *
 * does the actual mail transmission.
 *
 * called (only) by PalmToMail()
 */
int SendMail(struct Mail *mail, const char *from, const char* charset)
{
	FILE *sendmail_fd;
	char *mailstring;
	char date[MAILTMPLEN];

	rfc822_date(date);

	/* open up the sendmail command for writing */
	sendmail_fd = popen(gSendmail, "w");
	if (NULL == sendmail_fd) {
		/* this will NOT catch a bad gSendmail cmd. that can only be caught at
		 * pclose(). read the popen man page carefully. :(
		 */
		perror(gSendmail);
		return -1;
	}

	mailstring = (char *)malloc(1);
	mailstring[0] = '\0';
	mailstring = MailToString(mailstring, mail, from, date, charset);
	fprintf(sendmail_fd, "%s", mailstring);
	free(mailstring);

	return pclose(sendmail_fd);
}

/* DeleteMails()
 *
 * delete mails in Palm Mail app's "Deleted" category, also attempting to
 * delete the corresponding mails from the mail source.
 *
 * returns T if mails were deleted on the mailserver
 *
 * if gDeleteRead is set delete read mails on palm
 */
int
DeleteMails(MAILSTREAM *mailStream)
{
	int len, size, attr, deleteFlag = 0;
	int msgno;
	char buffer[0xffff];
	recordid_t id;
	listItem *searchResults=NULL;

	/* delete emails in the "Deleted" category on the Pilot */
	dlp_ResetDBIndex(gSocket, gDB);
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtgDeleted,
			buffer, &id, 0, &size, &attr))) {
		/* delete if we do SyncAll
		 * or if gDelete and not Hotsync-unread since
		 * the Hotsync-Unread Options says something of keeping
		 * all mails on the server
		 * with Hotsync-Seen we never come here anyway...
		 */
		if (gPrefs.syncType == mailSyncAll || 
		    (gDelete && gPrefs.syncType != mailSyncUnread)) {
			/* delete mail from mail source (but only if we're connected) */
			struct Mail mail;
			int multi=0; /* guard against one-to-many matches */

			unpack_Mail(&mail, buffer, len);

			searchResults = findMail(&mail, mailStream);
			while ((msgno = (int)pop(&searchResults))) {
				if (multi++ == 0){
					markDeleted(msgno, mailStream);
					deleteFlag = 1;
					infoOutput("X");
				}
				else
					infoOutput("while deleting, more than one msg matched on the server\n");
			}
		}
		/* no matter what, delete from Palm */
		dlp_DeleteRecord(gSocket, gDB, 0, id);
	}

	if (gDeleteRead) {
		dlp_ResetDBIndex(gSocket, gDB);
		while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtgInbox,
				buffer, &id, 0, &size, &attr))) {
			struct Mail mail;
			
			unpack_Mail(&mail, buffer, len);
			if (mail.read) {
				/* this mail has been read on the Palm; delete it */
				dlp_DeleteRecord(gSocket, gDB, 0, id);
			}
		}
	}
	
	/* delete archived/deleted records on Palm */
	dlp_CleanUpDatabase(gSocket, gDB); 
	return deleteFlag;
}

/* UpdateSeenFlags()
 *
 * Reads all the mails in the Inbox category and updates the flags
 * between the server and the palm.
 */
void
UpdateSeenFlags(MailCategory mailCtg, MAILSTREAM *mailStream)
{
	int len, size, attr;
	int msgno;
	char buffer[0xffff];
	recordid_t id;
	listItem *searchResults=NULL;

	/* consider every message in the Inbox */
	dlp_ResetDBIndex(gSocket,gDB);
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtg,
				buffer, &id, 0, &size, &attr))) {
		struct Mail mail;
		int multi=0; /* guard against one-to-many matches */

		unpack_Mail(&mail, buffer, len);

		/* try to find its conterpart in the mail source */
		searchResults = findMail(&mail, mailStream);
		while ((msgno = (int)pop(&searchResults))) {
			if (multi++ == 0) {
				if (mail.read) {
					markSeen(msgno, mailStream); /* mark source-side mail Seen */
					infoOutput("-");
				} else {
					if (isSeen(msgno, mailStream)) {
						mail.read = 1; /* mark Palm-side mail Seen */
						len = pack_Mail(&mail, buffer, sizeof(buffer));
						if (0 < dlp_WriteRecord(gSocket, gDB, 0, id, 0, 
									buffer, len, 0)) {
							infoOutput("+");
						} else {
							mm_log("  failure updating Palm\n", ERROR);
						}
					}
				}
			} else {
				infoOutput("while updating seen, more than one msg matched on the server\n");
			}
		}
	}
}

/* SynchronizeFlags()
 *
 * Reads all the mails in the Inbox category and updates the flags
 * between the server and the palm.
 *
 * Mails on the server which are already on the palm get marked
 * Mails on the palm, which aren't on the mailserver anymore are deleted on the palm
 *
 * based on UpdateSeenFlags, modified by jochen@pms.garcke.de
 */
void
SynchronizeFlags(MailCategory mailCtg, MAILSTREAM *mailStream)
{
	int len, size, attr;
	int msgno;
	char buffer[0xffff];
	recordid_t id;
	listItem *searchResults=NULL;

	/* consider every message in the Inbox */
	dlp_ResetDBIndex(gSocket,gDB);
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtg,
						     buffer, &id, 0, &size, &attr))) {
		struct Mail mail;
		int multi=0; /* guard against one-to-many matches */

		unpack_Mail(&mail, buffer, len);

		/* try to find its conterpart in the mail source */
		searchResults = findMail(&mail, mailStream);
		/* If no matching mail delete mail on palm */
		if (searchResults == NULL){
			infoOutput("d");
			dlp_DeleteRecord(gSocket, gDB, 0, id);		  
		}
		while ((msgno = (int)pop(&searchResults))) {
			if (multi++ == 0) {
				if (mail.read) {
					markSeenAndOnPalm(msgno, mailStream); /* mark source-side mail Seen */
					infoOutput("-");
				} else {
					if (isSeenMarkOnPalm(msgno, mailStream)) {
						mail.read = 1; /* mark Palm-side mail Seen */
						len = pack_Mail(&mail, buffer, sizeof(buffer));
						if (0 < dlp_WriteRecord(gSocket, gDB, 0, id, 0, 
									buffer, len, 0)) {
							infoOutput("+");
						} else {
							mm_log("  failure updating Palm\n", ERROR);
						}
					}
				}
			} else {
				infoOutput("while updating seen, more than one msg matched on the server\n");
			}
		}
	}
}
